import django
from django.conf import settings


def pytest_configure():
    """Initialize Django settings."""
    settings.configure(
        SECRET_KEY="secret",
        DEBUG=True,
        INSTALLED_APPS=[
            "django.contrib.admin",
            "django.contrib.auth",
            "django.contrib.contenttypes",
            "django.contrib.sessions",
            "django.contrib.messages",
            "django.contrib.staticfiles",
            "tests",
            "worf",
        ],
        MIDDLEWARE=[
            "django.middleware.common.CommonMiddleware",
            "django.contrib.sessions.middleware.SessionMiddleware",
            "django.contrib.auth.middleware.AuthenticationMiddleware",
            "django.contrib.messages.middleware.MessageMiddleware",
        ],
        ROOT_URLCONF="tests.urls",
        DATABASES={
            "default": {
                "ENGINE": "django.db.backends.sqlite3",
                "NAME": "db.sqlite3",
            }
        },
        PASSWORD_HASHERS=["django.contrib.auth.hashers.MD5PasswordHasher"],
        TIME_ZONE="UTC",
        USE_I18N=True,
        USE_L10N=True,
        USE_TZ=True,
        STATIC_URL="/static/",
    )

    django.setup()



class Serializer:
    def create(self):
        return []

    def read(self, model):
        return {}

    def write(self):
        return []


class LegacySerializer(Serializer):
    def __init__(self, model_class, api_method):
        self.api_method = api_method
        self.model_class = model_class

    def read(self, model):
        payload = getattr(model, self.api_method)()
        if not isinstance(payload, dict):
            msg = f"{model.__name__}.{self.api_method}() did not return a dictionary"
            raise ImproperlyConfigured(msg)
        return payload

    def write(self):
        return getattr(self.model_class(), f"{self.api_method}_update_fields")()



from django.core.exceptions import ValidationError

from worf.casing import snake_to_camel
from worf.views.base import AbstractBaseAPI


class CreateAPI(AbstractBaseAPI):
    def serialize(self):
        return {}

    def post(self, request, *args, **kwargs):
        serializer = self.get_serializer()

        create_fields = serializer.create()
        for key in self.bundle.keys():
            self.validate_bundle(key)
            # ignore create_fields for now if it's empty
            # this should be moved into validate bundle
            if create_fields and key not in create_fields:
                raise ValidationError(
                    "{} not allowed when creating {}".format(
                        snake_to_camel(key), self.name
                    )
                )

        new_instance = self.model.objects.create(**self.bundle)
        return self.render_to_response(serializer.read(new_instance), 201)



def transform_to_dict(choices):
    """Convert various choices list of lists into json object."""
    # TODO this is copied from context_processors.
    transformed_choices = dict()
    for each in choices:
        transformed_choices[each[0]] = each[1]
    return transformed_choices



from worf.views.detail import (
    DetailAPI,
    DetailUpdateAPI,
)
from worf.views.list import (
    ListAPI,
    ListCreateAPI,
)
from worf.views.base import (
    APIResponse,
    AbstractBaseAPI,
)
from worf.views.create import CreateAPI



from worf.testing import deserialize


def test_user_get(client, django_user_model):
    user = django_user_model.objects.create_user(username="test", password="password")
    response = client.get("/user/{}/".format(user.pk))
    assert response.status_code == 200
    assert deserialize(response)["username"] == "test"

    bundle = dict(username="testtest", email="something@example.com")
    response = client.patch(
        "/user/{}/".format(user.pk), bundle, content_type="application/json"
    )
    assert response.status_code == 200
    assert deserialize(response)["username"] == "testtest"
    assert deserialize(response)["email"] == "something@example.com"



from django.urls import include, path

from tests.views import DummyAPI, UserAPI

urlpatterns = [
    path("", DummyAPI.as_view()),
    path("<str:id>/", DummyAPI.as_view()),
    path("user/<str:id>/", UserAPI.as_view()),
]



from datetime import datetime
from uuid import UUID

from django.conf import settings
from django.core.exceptions import ValidationError
from django.core.validators import validate_email
from django.utils.html import strip_tags

from worf.exceptions import NotImplementedInWorfYet
from worf.casing import snake_to_camel


class ValidationMixin:
    boolean_values = {
        "1": True,
        "0": False,
        "true": True,
        "false": False,
    }

    secure_fields = [
        "password",
        "password_confirmation",
    ]

    def _coerce_bool(self, value):
        if isinstance(value, bool):
            return value

        return self.boolean_values.get(str(value).lower().strip(), None)

    def _validate_boolean(self, key):
        value = self.bundle[key]
        coerced = self._coerce_bool(value)

        if not isinstance(coerced, bool):
            raise ValidationError(
                f"Field {snake_to_camel(key)} accepts a boolean, got {value}, coerced to {coerced}"
            )

        return coerced

    def _validate_date(self, key):
        value = self.bundle[key]

        if not isinstance(value, str):
            return None

        return datetime.strptime(value, "%Y-%m-%d")

    def _validate_many_to_many(self, key):
        value = self.bundle[key]

        if not isinstance(value, list):
            raise ValidationError(
                f"Field {snake_to_camel(key)} accepts an array, got {type(value)} {value}"
            )

        self.coerce_array_of_integers(key)

    def _validate_string(self, key, max_length):
        value = self.bundle[key]

        if not isinstance(value, str):
            raise ValidationError(f"Field {snake_to_camel(key)} accepts string")

        if key not in self.secure_fields:
            value = strip_tags(value).strip()

        if max_length is not None and len(value) > max_length:
            raise ValidationError(
                f"Field {snake_to_camel(key)} accepts a maximum of {max_length} characters"
            )

        return value

    def _validate_int(self, key):
        value = self.bundle[key]

        if value is None or value == "":
            return None

        try:
            integer = int(value)
        except (TypeError, ValueError):
            raise ValidationError(f"Field {snake_to_camel(key)} accepts an integer")

        return integer

    def _validate_positive_int(self, key):
        integer = self._validate_int(key)

        if integer < 0:
            raise ValidationError(
                f"Field {snake_to_camel(key)} accepts a positive integer"
            )

        return integer

    ############################################################################
    # Public methods for use downstream
    def coerce_array_of_integers(self, key):
        try:
            self.bundle[key] = [int(id) for id in self.bundle[key]]
        except ValueError:
            message = f"Field {snake_to_camel(key)} accepts an array of integers. Got {self.bundle[key]} instead."
            raise ValidationError(message + " I couldn't coerce the values.")

    def get_field_type(self, key):
        return self.model._meta.get_field(key).get_internal_type()

    def validate_int(self, value):
        if not isinstance(value, int):
            raise ValidationError(f"Expected integer, got {value}")

    def validate_uuid(self, value):
        if value is None:
            raise ValidationError(f"Expected UUID, got {value}")
        try:
            return UUID(value)
        except (TypeError, ValueError):
            raise ValidationError(f"Expected UUID, got {value}")

    def validate_email(self, value):
        if not isinstance(value, str):
            raise ValidationError(f"{value} is not a valid email address")

        email = value.strip().lower()
        try:
            validate_email(email)
        except ValidationError:
            raise ValidationError(f"{value} is not a valid email address")
        return email

    def validate_bundle(self, key):
        """
        Handle basic type validation and coercion.

        @param key _str_ the model attribute to check against.
        @return value:
            - If the HTTP method is not PATCH and `key` does not exist in the
        model serializer method, return False.
            - If an error is detected, ValidationError will be raised
            - If all checks pass, True is returned

        Side Effects:
        As various bundle objects are parsed and validated, we reset the bundle.
        This may result in self.bundle changes.

        We expect to set a fully validated bundle keys and values.
        """
        # POST cannot validate b/c it allows fields not in api_update_fields
        # self.request.method == 'POST' or
        serializer = self.get_serializer()
        if self.request.method in ("PATCH", "PUT") and key not in serializer.write():
            message = f"{snake_to_camel(key)} is not editable"
            if settings.DEBUG:
                message += f":: {serializer}"
            raise ValidationError(message)

        if not hasattr(self.model, key):
            raise ValidationError(f"{snake_to_camel(key)} does not exist")

        field_type = self.get_field_type(key)

        if hasattr(self, f"validate_{key}"):
            self.bundle[key] = getattr(self, f"validate_{key}")(self.bundle[key])

        elif field_type in ["CharField", "TextField", "SlugField"]:
            max_length = self.model._meta.get_field(key).max_length
            self.bundle[key] = self._validate_string(key, max_length)

        elif field_type == "EmailField":
            self.bundle[key] = self.validate_email(self.bundle[key])

        elif field_type in ["IntegerField", "SmallIntegerField"]:
            # TODO check size of SmallIntegerField
            self.bundle[key] = self._validate_int(key)

        elif field_type == "PositiveIntegerField":
            self.bundle[key] = self._validate_positive_int(key)

        elif field_type == "ManyToManyField":
            self._validate_many_to_many(key)

        elif field_type in ["BooleanField"]:
            self.bundle[key] = self._validate_boolean(key)

        elif field_type in ["ForeignKey"]:
            pass  # Django will raise an exception if handled improperly

        elif field_type in ["DateField"]:
            self.bundle[key] = self._validate_date(key)

        elif field_type in ["JSONField"]:
            pass
            # TODO check the type of json object we're expecting
            # try:
            #     json.loads(self.bundle[key])
            # except ValueError:
            #     raise ValidationError(f"Field {snake_to_camel(key)} requires valid JSON")

        else:
            message = f"{field_type} has no validation method for {key}"
            if settings.DEBUG:
                message += f":: Received {self.bundle[key]}"
            raise NotImplementedInWorfYet(message)
            # TODO
            # FileField
            # UUIDField
            # We don't handle 1:1 or FKs.

        return True



from worf.exceptions import NamingThingsError


def snake_to_camel(snake):
    invalid_msg = f"{snake} is not valid snake case. "
    if snake.lower() != snake:
        raise NamingThingsError(invalid_msg + "It has capital letters!")

    components = snake.split("_")
    if not "".join(x for x in components).isalpha():
        raise NamingThingsError(invalid_msg + "It has special chars!")

    return components[0] + "".join(x.title() for x in components[1:])


def camel_to_snake(camel):
    if camel.lower() == camel:
        return camel

    invalid_msg = f"{camel} is not valid camel case. "
    if not camel.isalpha():
        raise NamingThingsError(invalid_msg + "It has non alphabetical chars!")

    snake = ""
    last_was_upper = True
    for value in camel:
        if value.isupper():
            if last_was_upper:
                raise NamingThingsError(
                    invalid_msg + "It has multiple upper-case chars sequentially!"
                )
            snake += "_" + value.lower()
        else:
            snake += value

        last_was_upper = value.isupper()
    return snake


def whitespace_to_camel(string):
    pos = string.find(" ")
    if pos == -1:
        return string[:1].lower() + string[1:]

    new_string = string[:pos] + string[pos + 1 :].capitalize()
    return whitespace_to_camel(new_string)
